import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client
)


def test_permissions(tornado_server):
    admin_client = Client(tornado_server, "admin")
    user_client = Client(tornado_server, "user")

    admin_client.create("/sites", name="Test Site")
    admin_client.create("/sites/1/attributes", resource_name="Network", name="attr1")
    admin_client.create("/sites/1/networks", cidr="10.0.0.0/24")

    user_client.get("/")  # Just create user id 2

    assert_success(
        user_client.get("/users/1/permissions"),
        {"permissions": {"1": {"permissions": ["admin"], "site_id": 1, "user_id": 1}}}
    )

    assert_success(
        admin_client.get("/users/1/permissions/1"),
        {"permission": {"permissions": ["admin"], "site_id": 1, "user_id": 1}}
    )

    # Explicitly Set permissions for user_client to have none.
    assert_success(admin_client.update(
        "/users/2/permissions/1", permissions=[]
    ),
        {"permission": {"permissions": [], "site_id": 1, "user_id": 2}}
    )

    # User shouldn't be able to update site or create/update other resources
    assert_error(
        user_client.update("/sites/1", name="attr1"),
        403
    )
    assert_error(
        user_client.create("/sites/1/attributes", name="attr2"),
        403
    )
    assert_error(
        user_client.update("/sites/1/attributes/1", required=True),
        403
    )
    assert_error(
        user_client.create("/sites/1/networks", cidr="10.0.0.0/8"),
        403
    )
    assert_error(
        user_client.update("/sites/1/networks/1", attributes={"attr1": "foo"}),
        403
    )



import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app, user, session, auth_token
from .util import (
    assert_error, assert_success, Client
)

def test_user_with_secret_key(tornado_server):
    user1_client = Client(tornado_server, "user1")
    user2_client = Client(tornado_server, "user2")

    # Small requests to make user accounts in order.
    user1_client.get("/users/0")
    user2_client.get("/users/0")

    assert_success(
        user1_client.get("/users/0"),
        {"user": {"email": "user1@localhost", "id": 1, "permissions": {}}}
    )

    response = user1_client.get("/users/0?with_secret_key")
    assert_success(
        response,
        {"user": {
            "email": "user1@localhost",
            "id": 1,
            "secret_key": response.json()["data"]["user"]["secret_key"],
            "permissions": {}}
        }
    )

    response = user1_client.get("/users/1?with_secret_key")
    assert_success(
        response,
        {"user": {
            "email": "user1@localhost",
            "id": 1,
            "secret_key": response.json()["data"]["user"]["secret_key"],
            "permissions": {}}
        }
    )

    response = user1_client.get("/users/2?with_secret_key")
    assert_error(response, 403)


def test_user_rotate_secret_key(tornado_server):
    user1_client = Client(tornado_server, "user1")
    user2_client = Client(tornado_server, "user2")

    # Small requests to make user accounts in order.
    user1_client.get("/users/0")
    user2_client.get("/users/0")

    assert_success(user1_client.post("/users/1/rotate_secret_key"))
    assert_error(user1_client.post("/users/2/rotate_secret_key"), 403)



from nsot import models

from .fixtures import session, site, user, admin


def test_reparent_bug_issues_27(session, admin, site):
    """
        Test for bug described at https://github.com/dropbox/nsot/issues/27
    """

    net_8  = models.Network.create(session, admin.id, site.id, u"10.0.0.0/8")
    net_31 = models.Network.create(session, admin.id, site.id, u"10.17.244.128/31")
    net_25 = models.Network.create(session, admin.id, site.id, u"10.16.1.0/25")

    assert net_8.parent_id is None
    assert net_31.parent_id == net_8.id
    assert net_25.parent_id == net_8.id



from __future__ import unicode_literals

from calendar import timegm
from cryptography.fernet import (Fernet, InvalidToken)
from datetime import datetime
from email.utils import parseaddr
from operator import attrgetter
import functools
import ipaddress
import re
import json
import logging
import time

from sqlalchemy import create_engine, or_, union_all, desc
from sqlalchemy.event import listen
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, object_session, aliased, validates
from sqlalchemy.orm import synonym, sessionmaker, Session as _Session, backref
from sqlalchemy.schema import Column, ForeignKey, Index
from sqlalchemy.sql import func, label, literal, false
from sqlalchemy.types import Integer, String, Text, Boolean, SmallInteger
from sqlalchemy.types import Enum, DateTime, VARBINARY

from . import constants
from . import exc
from .permissions import PermissionsFlag
from .settings import settings


log = logging.getLogger(__name__)

RESOURCE_BY_IDX = (
    "Site", "Network", "Attribute", "Permission", "Device"
)
RESOURCE_BY_NAME = {
    obj_type: idx
    for idx, obj_type in enumerate(RESOURCE_BY_IDX)
}

CHANGE_EVENTS = ("Create", "Update", "Delete")
IP_VERSIONS = ("4", "6")


VALID_CHANGE_RESOURCES = set(RESOURCE_BY_IDX)
VALID_ATTRIBUTE_RESOURCES = set([
    "Network", "Device",
])


class Session(_Session):
    """ Custom session meant to utilize add on the model.

        This Session overrides the add/add_all methods to prevent them
        from being used. This is to for using the add methods on the
        models themselves where overriding is available.
    """

    _add = _Session.add
    _add_all = _Session.add_all
    _delete = _Session.delete

    def add(self, *args, **kwargs):
        raise NotImplementedError("Use add method on models instead.")

    def add_all(self, *args, **kwargs):
        raise NotImplementedError("Use add method on models instead.")

    def delete(self, *args, **kwargs):
        raise NotImplementedError("Use delete method on models instead.")


Session = sessionmaker(class_=Session)

class Model(object):
    """ Custom model mixin with helper methods. """

    def __repr__(self):
        cls_name = self.__class__.__name__
        return u'%s(id=%r)' % (cls_name, self.id)

    @property
    def session(self):
        return object_session(self)

    @classmethod
    def query(cls):
        """Return a Query using session defaults."""
        # The Model.query() method doesn't work on classmethods. We need a
        # scoped_session for that
        from .meta import ScopedSession
        return ScopedSession().query(cls)

    @classmethod
    def get_or_create(cls, session, **kwargs):
        instance = session.query(cls).filter_by(**kwargs).scalar()
        if instance:
            return instance, False

        instance = cls(**kwargs)
        instance.add(session)

        return instance, True

    @property
    def model_name(self):
        obj_name = type(self).__name__
        return obj_name

    @classmethod
    def before_create(cls, session, user_id):
        """ Hook for before object creation."""

    def after_create(self, user_id):
        """ Hook for after object creation."""

    @classmethod
    def create(cls, session, _user_id, **kwargs):
        commit = kwargs.pop("commit", True)
        try:
            cls.before_create(session, _user_id)
            obj = cls(**kwargs).add(session)
            session.flush()
            obj.after_create(_user_id)
            Change.create(session, _user_id, "Create", obj)
            if commit:
                session.commit()
        except Exception:
            session.rollback()
            raise

        return obj

    def update(self, user_id, **kwargs):
        session = self.session
        try:
            for key, value in kwargs.iteritems():
                setattr(self, key, value)

            session.flush()
            Change.create(session, user_id, "Update", self)
            session.commit()
        except Exception:
            session.rollback()
            raise

    def add(self, session):
        session._add(self)
        return self

    def before_delete(self):
        """ Hook for extra model cleanup before delete. """

    def after_delete(self):
        """ Hook for extra model cleanup after delete. """

    def delete(self, user_id):
        session = self.session
        try:
            Change.create(session, user_id, "Delete", self)
            self.before_delete()
            session._delete(self)
            self.after_delete()
            session.commit()
        except Exception:
            session.rollback()
            raise


Model = declarative_base(cls=Model)


# Foreign Keys are ignored by default in SQLite. Don't do that.
def _set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON;")
    cursor.close()


def get_db_engine(url, echo=False):
    engine = create_engine(url, pool_recycle=300, echo=echo)
    if engine.driver == "pysqlite":
        listen(engine, "connect", _set_sqlite_pragma)

    return engine


def get_db_session(db_engine=None, database=None):
    """
    Return a usable session object.

    If not provided, this will attempt to retreive the database and db_engine
    from settings. If settings have not been updated from a config, this will
    return ``None``.

    :param db_engine:
        Database engine to use

    :param database:
        URI for database
    """
    if database is None:
        database = settings.database
        if database is None:
            return None
    if db_engine is None:
        db_engine = get_db_engine(database)
    Session.configure(bind=db_engine)
    return Session()


def flush_transaction(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        dryrun = kwargs.pop("dryrun", False)
        try:
            ret = method(self, *args, **kwargs)
            if dryrun:
                self.session.rollback()
            else:
                self.session.flush()
        except Exception:
            log.exception("Transaction Failed. Rolling back.")
            if self.session is not None:
                self.session.rollback()
            raise
        return ret
    return wrapper


class User(Model):

    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    email = Column(String(length=255), unique=True, nullable=False)

    # The user's secret key is used to generate their auth_token
    secret_key = Column(String(length=255), default=Fernet.generate_key)

    @validates("email")
    def validate_email(self, key, value):
        _, email = parseaddr(value)
        if email == '' or "@" not in value:
            raise exc.ValidationError("Must contain a valid e-mail address")
        return value

    def to_dict(self, with_permissions=False, with_secret_key=False):
        out = {
            "id": self.id,
            "email": self.email,
        }

        if with_secret_key:
            out["secret_key"] = self.secret_key

        if with_permissions:
            out["permissions"] = self.get_permissions()

        return out

    def rotate_secret_key(self):
        self.secret_key = Fernet.generate_key()

    def get_permissions(self, site_id=None):
        query = self.session.query(Permission).filter_by(
            user_id=self.id
        )

        if site_id is not None:
            query = query.filter_by(site_id=site_id)

        permissions = query.all()

        return {
            # JSON keys can't be ints so be consistent in
            # python
            str(permission.site_id): permission.to_dict()
            for permission in permissions
        }

    def generate_auth_token(self):
        """Serialize user data and encrypt token."""
        # Serialize data
        data = json.dumps({'email': self.email })

        # Encrypt w/ servers's secret_key
        f = Fernet(bytes(settings.secret_key))
        auth_token = f.encrypt(bytes(data))
        return auth_token

    def verify_secret_key(self, secret_key):
        """Validate secret_key"""
        return secret_key == self.secret_key

    @classmethod
    def verify_auth_token(cls, email, auth_token,
                          expiration=None, session=None):
        """Verify token and return a User object."""
        if expiration is None:
            expiration = settings.auth_token_expiry

        # First we lookup the user by email
        if session is None:
            query = User.query()
        else:
            query = session.query(User)
        user = query.filter_by(email=email).scalar()

        if user is None:
            log.debug('Invalid user when verifying token')
            return None  # Invalid user

        # Decrypt auth_token w/ user's secret_key
        f = Fernet(bytes(settings.secret_key))
        try:
            decrypted_data = f.decrypt(bytes(auth_token), ttl=expiration)
        except InvalidToken:
            log.debug('Invalid/expired auth_token when decrypting.')
            return None  # Invalid token

        # Deserialize data
        try:
            data = json.loads(decrypted_data)
        except ValueError:
            log.debug('Token could not be deserialized.')
            return None  # Valid token, but expired

        if email != data['email']:
            log.debug('Invalid user when deserializing.')
            return None  # User email did not match payload
        return user


class Permission(Model):
    __tablename__ = "permissions"
    __table_args__ = (
        Index("site_user_idx", "site_id", "user_id", unique=True),
    )

    id = Column(Integer, primary_key=True)
    site_id = Column(Integer, ForeignKey("sites.id", ondelete="CASCADE"), nullable=False)
    user_id = Column(Integer, ForeignKey("users.id"), nullable=False)

    permissions_flag = Column(Integer, default=0, nullable=False)

    @property
    def permissions(self):
        perms = []
        flag = PermissionsFlag(self.permissions_flag)
        for word in PermissionsFlag.words:
            if flag.has(word):
                perms.append(word)
        return perms

    @permissions.setter
    def permissions(self, value):
        flag = PermissionsFlag()
        for word in value:
            flag.set(word)
        self.permissions_flag = flag.dump()

    def to_dict(self):
        return {
            # Not including primary key here since all operations
            # will be on the composite key of site_id and user_id.
            "site_id": self.site_id,
            "user_id": self.user_id,
            "permissions": self.permissions,
        }


class Site(Model):
    """ A namespace for subnets, ipaddresses, attributes. """

    __tablename__ = "sites"

    id = Column(Integer, primary_key=True)
    name = Column(String(length=255), unique=True, nullable=False)
    description = Column(Text, default="", nullable=False)

    # All generic resources are expected to have a site_id attribute.
    site_id = synonym("id")


    def after_create(self, user_id):
        Permission.create(
            self.session, user_id, site_id=self.id, user_id=user_id,
            permissions=["admin"]
        )

    def devices(self, attribute_name=None, attribute_value=None):
        """ Helper method for grabbing Networks.

            Args:
                attribute_name: Filter to networks that contain this attribute name
                attribute_value: Filter to networks that contain this attribute value
        """

        if attribute_value is not None and attribute_name is None:
            raise ValueError("attribute_value requires attribute_name to be set.")

        query = self.session.query(Device)

        if attribute_name is not None:
            query = query.outerjoin(Device.attr_idx).filter(
                DeviceAttributeIndex.name == attribute_name
            )
            if attribute_value is not None:
                query = query.filter(DeviceAttributeIndex.value == attribute_value)

        query = query.filter(Device.site_id==self.id)

        return query

    def networks(self, include_networks=True, include_ips=False, root=False,
                 subnets_of=None, supernets_of=None,
                 attribute_name=None, attribute_value=None):
        """ Helper method for grabbing Networks.

            Args:
                include_networks: Whether the response should include non-ip
                                  address networks
                include_ips: Whether the response should include ip addresses
                root: Only return networks at the root.
                subnets_of: Only return subnets of the given CIDR
                supernets_of: Only return supernets of the given CIDR
                attribute_name: Filter to networks that contain this attribute name
                attribute_value: Filter to networks that contain this attribute value
        """

        if not any([include_networks, include_ips]):
            return self.session.query(Network).filter(false())

        if attribute_value is not None and attribute_name is None:
            raise ValueError("attribute_value requires attribute_name to be set.")

        if all([subnets_of, supernets_of]):
            raise ValueError("subnets_of and supernets_of are mutually exclusive.")

        query = self.session.query(Network)

        if attribute_name is not None:
            query = query.outerjoin(Network.attr_idx).filter(
                NetworkAttributeIndex.name == attribute_name
            )
            if attribute_value is not None:
                query = query.filter(NetworkAttributeIndex.value == attribute_value)

        query = query.filter(Network.site_id==self.id)

        if not all([include_networks, include_ips]):
            if include_networks:
                query = query.filter(Network.is_ip == False)
            if include_ips:
                query = query.filter(Network.is_ip == True)

        if root:
            query = query.filter(Network.parent_id == None)

        if subnets_of is not None:
            subnets_of = ipaddress.ip_network(unicode(subnets_of))
            query = query.filter(
                Network.ip_version == str(subnets_of.version),
                Network.prefix_length > subnets_of.prefixlen,
                Network.network_address >= subnets_of.network_address.packed,
                Network.broadcast_address <= subnets_of.broadcast_address.packed
            )

        if supernets_of is not None:
            supernets_of = ipaddress.ip_network(unicode(supernets_of))
            query = query.filter(
                Network.ip_version == str(supernets_of.version),
                Network.prefix_length < supernets_of.prefixlen,
                Network.network_address <= supernets_of.network_address.packed,
                Network.broadcast_address >= supernets_of.broadcast_address.packed
            )

        return query


    @validates("name")
    def validate_name(self, key, value):
        if not value:
            raise exc.ValidationError("Name is a required field.")
        return value


    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
        }


class AttributeIndexMixin(object):
    """ Shared columns for AttributeIndex tables. """
    id = Column(Integer, primary_key=True)

    name = Column(String(length=64), nullable=False, index=True)
    value = Column(String(length=255), nullable=False, index=True)

    @declared_attr
    def attribute_id(self):
        return Column(Integer, ForeignKey("attributes.id"), nullable=False)

    @declared_attr
    def attribute(self):
        return relationship("Attribute")


class NetworkAttributeIndex(AttributeIndexMixin, Model):
    """ An Inverted Index for looking up Networks by their attributes."""

    __tablename__ = "network_attribute_index"

    resource_id = Column(Integer, ForeignKey("networks.id"), nullable=False)
    resource = relationship("Network", backref="attr_idx")


class DeviceAttributeIndex(AttributeIndexMixin, Model):
    """ An Inverted Index for looking up Devices by their attributes."""

    __tablename__ = "device_attribute_index"

    resource_id = Column(Integer, ForeignKey("devices.id"), nullable=False)
    resource = relationship("Device", backref="attr_idx")


class AttributeModelMixin(object):
    """ Shared columns and methods for resources that support attributes. """

    # Attributes is a Serialized LOB field. Lookups of these attributes
    # is done against an Inverted Index table
    _attributes = Column("attributes", Text)

    def _purge_attribute_index(self):
        index_table = self._index_class.__table__
        self.session.execute(
            index_table.delete().where(index_table.c.resource_id == self.id)
        )

    def get_attributes(self):
        if self._attributes is None:
            return {}
        return json.loads(self._attributes)

    def set_attributes(self, attributes, valid_attributes=None):
        if not isinstance(attributes, dict):
            raise exc.ValidationError("Expected dictionary but received {}".format(
                type(attributes)
            ))

        if valid_attributes is None:
            valid_attributes = Attribute.all_by_name(self.session, self._resource_name)

        missing_attributes = {
            attribute.name
            for attribute in valid_attributes.itervalues()
            if attribute.required and attribute.name not in attributes
        }

        if missing_attributes:
            raise exc.ValidationError("Missing required attributes: {}".format(
                ", ".join(missing_attributes)
            ))

        inserts = []
        for name, value in attributes.iteritems():
            if name not in valid_attributes:
                raise exc.ValidationError("Attribute name ({}) does not exist.".format(
                    name
                ))

            if not isinstance(name, basestring):
                raise exc.ValidationError("Attribute names must be a string type")

            attribute = valid_attributes[name]
            inserts.extend(attribute.validate_value(value))

        for insert in inserts:
            insert["resource_id"] = self.id

        self._purge_attribute_index()
        if inserts:
            index_table = self._index_class.__table__
            self.session.execute(index_table.insert(), inserts)

        self._attributes = json.dumps(attributes)

    def update(self, user_id, **kwargs):
        session = self.session
        try:
            for key, value in kwargs.iteritems():
                if key == "attributes":
                    self.set_attributes(value)
                else:
                    setattr(self, key, value)
            session.flush()
            Change.create(session, user_id, "Update", self)
            session.commit()
        except Exception:
            session.rollback()
            raise


class Device(AttributeModelMixin, Model):
    """ Represents a network device. """

    # Class attributes used by Mixin
    _index_class = DeviceAttributeIndex
    _resource_name = "Device"

    __tablename__ = "devices"
    __table_args__ = (
        Index(
            "hostname_idx",
            "site_id", "hostname",
            unique=True
        ),
    )

    id = Column(Integer, primary_key=True)
    site_id = Column(Integer, ForeignKey("sites.id"), nullable=False, index=True)
    hostname = Column(String(255), nullable=False, index=True)

    @classmethod
    def create(cls, session, user_id, site_id, hostname, attributes=None, **kwargs):

        # TODO(gary): Decent duplication between Device and  Network create method.
        #             Clean this up at some point.
        commit = kwargs.pop("commit", True)
        if attributes is None:
            attributes = {}

        try:
            obj = cls(site_id=site_id, hostname=hostname)
            obj.add(session)
            # Need to get a primary key for the new network to update subnets.
            session.flush()
            # Attributes have to be added after the initial flush since we need the
            # id to setup the inverted index. This will mean we have to send another
            # update later. This could be improved by separating the index step but
            # probably isn't worth it.
            obj.set_attributes(attributes)
            Change.create(session, user_id, "Create", obj)

            if commit:
                session.commit()

        except Exception:
            session.rollback()
            raise

        return obj

    @validates("hostname")
    def validate_hostname(self, key, value):
        if not value:
            raise exc.ValidationError("hostname must be non-zero length string.")
        return value

    def to_dict(self):
        return {
            "id": self.id,
            "site_id": self.site_id,
            "hostname": self.hostname,
            "attributes": self.get_attributes(),
        }


class Network(AttributeModelMixin, Model):
    """ Represents a subnet or ipaddress. """

    # Class attributes used by Mixin
    _index_class = NetworkAttributeIndex
    _resource_name = "Network"

    __tablename__ = "networks"
    __table_args__ = (
        Index(
            "cidr_idx",
            "site_id", "ip_version", "network_address", "prefix_length",
            unique=True
        ),
    )

    id = Column(Integer, primary_key=True)
    site_id = Column(Integer, ForeignKey("sites.id"), nullable=False, index=True)

    ip_version = Column(String(1), nullable=False, index=True)

    # Root networks will be NULL while other networks will point to
    # their supernet.
    parent_id = Column(Integer, ForeignKey("networks.id"), nullable=True)

    network_address = Column(VARBINARY(16), nullable=False, index=True)
    # While derivable from network/prefix this is useful as it enables
    # easy querys of the nested set variety.
    broadcast_address = Column(VARBINARY(16), nullable=False, index=True)

    prefix_length = Column(Integer, nullable=False, index=True)

    # Simple boolean
    is_ip = Column(Boolean, nullable=False, default=False, index=True)

    @validates("ip_version")
    def validate_ip_version(self, key, value):
        if value not in IP_VERSIONS:
            raise exc.ValidationError("Invalid IP Version.")
        return value

    def to_dict(self):
        network_address = ipaddress.ip_address(self.network_address)

        return {
            "id": self.id,
            "parent_id": self.parent_id,
            "site_id": self.site_id,
            "is_ip": self.is_ip,
            "ip_version": self.ip_version,
            "network_address": network_address.exploded,
            "prefix_length": self.prefix_length,
            "attributes": self.get_attributes(),
        }

    def before_delete(self):
        self._purge_attribute_index()

    def supernets(self, session, direct=False, discover_mode=False, for_update=False):
        """ Get networks that are a supernet of a network.

            Args:
                direct: Only return direct supernet.
                discover_mode: Prevent new networks from bailing for missing parent_id
                for_update: Lock these rows because they're selected for updating.

        """

        if self.parent_id is None and not discover_mode:
            return self.session.query(Network).filter(false())

        if discover_mode and direct:
            raise ValueError("direct is incompatible with discover_mode")

        query = session.query(Network)
        if for_update:
            query = query.with_for_update()

        if direct:
            return query.filter(Network.id == self.parent_id)

        return query.filter(
            Network.is_ip == False,
            Network.ip_version == self.ip_version,
            Network.prefix_length < self.prefix_length,
            Network.network_address <= self.network_address,
            Network.broadcast_address >= self.broadcast_address
        )

    def subnets(self, session, include_networks=True, include_ips=False,
                direct=False, for_update=False):
        """ Get networks that are subnets of a network.

            Args:
                include_networks: Whether the response should include non-ip address networks
                include_ips: Whether the response should include ip addresses
                direct: Only return direct subnets.
                for_update: Lock these rows because they're selected for updating.
        """

        if not any([include_networks, include_ips]) or self.is_ip:
            return self.session.query(Network).filter(false())

        query = session.query(Network)
        if for_update:
            query = query.with_for_update()

        if not all([include_networks, include_ips]):
            if include_networks:
                query = query.filter(Network.is_ip == False)
            if include_ips:
                query = query.filter(Network.is_ip == True)

        if direct:
            return query.filter(Network.parent_id == self.id)

        return query.filter(
            Network.site_id == self.site_id,
            Network.ip_version == self.ip_version,
            Network.prefix_length > self.prefix_length,
            Network.network_address >= self.network_address,
            Network.broadcast_address <= self.broadcast_address
        )

    @property
    def cidr(self):
        return u"{}/{}".format(
            ipaddress.ip_address(self.network_address).exploded,
            self.prefix_length
        )
    @property
    def ip_network(self):
        return ipaddress.ip_network(self.cidr)

    def __repr__(self):
        return "Network<{}>".format(self.cidr)

    def reparent_subnets(self, session):
        query = session.query(Network).with_for_update().filter(
            Network.parent_id == self.parent_id,
            Network.id != self.id,  # Don't include yourself...
            Network.prefix_length > self.prefix_length,
            Network.ip_version == self.ip_version,
            Network.network_address >= self.network_address,
            Network.broadcast_address <= self.broadcast_address
        )

        query.update({Network.parent_id: self.id})

    @classmethod
    def create(cls, session, user_id, site_id, cidr, attributes=None, **kwargs):
        commit = kwargs.pop("commit", True)
        if attributes is None:
            attributes = {}
        if not cidr:
            msg = "Invalid CIDR: {}. Must be IPv4/IPv6 notation.".format(cidr)
            raise exc.ValidationError(msg)

        network = cidr
        if isinstance(cidr, unicode):
            network = ipaddress.ip_network(cidr)

        is_ip = False
        if network.network_address == network.broadcast_address:
            is_ip = True

        kwargs = {
            "site_id": site_id,
            "ip_version": str(network.version),
            "network_address": network.network_address.packed,
            "broadcast_address": network.broadcast_address.packed,
            "prefix_length": network.prefixlen,
            "is_ip": is_ip,
        }

        try:
            obj = cls(**kwargs)
            obj.add(session)
            # Need to get a primary key for the new network to update subnets.
            session.flush()
            # Attributes have to be added after the initial flush since we need the
            # id to setup the inverted index. This will mean we have to send another
            # update later. This could be improved by separating the index step but
            # probably isn't worth it.
            obj.set_attributes(attributes)

            supernets = obj.supernets(session, discover_mode=True, for_update=True).all()
            if supernets:
                parent = max(supernets, key=attrgetter("prefix_length"))
                obj.parent_id = parent.id

            if obj.parent_id is None and is_ip:
                raise exc.ValidationError("IP Address needs base network.")

            if not is_ip:
                obj.reparent_subnets(session)

            session.flush()
            Change.create(session, user_id, "Create", obj)

            if commit:
                session.commit()
        except Exception:
            session.rollback()
            raise

        return obj

class Attribute(Model):

    __tablename__ = "attributes"
    __table_args__ = (
        Index(
            "name_idx",
            "site_id", "resource_name", "name",
            unique=True
        ),
    )

    id = Column(Integer, primary_key=True)

    site_id = Column(Integer, ForeignKey("sites.id"), nullable=False)

    resource_name = Column(String(20), nullable=False, index=True)

    # This is purposely not unique as there is a compound index with site_id.
    name = Column(String(length=64), nullable=False)
    description = Column(Text, default="", nullable=False)

    # The resource must contain a key and value
    required = Column(Boolean, default=False, nullable=False)

    # In UIs this attribute will be displayed by default.
    # required implies display.
    display = Column(Boolean, default=False, nullable=False)

    # Attribute values are expected as lists of strings.
    multi = Column(Boolean, default=False, nullable=False)

    _constraints = Column("constraints", Text, nullable=True)

    @validates("name")
    def validate_name(self, key, value):
        if not value:
            raise exc.ValidationError("Name is a required field.")

        if not constants.ATTRIBUTE_NAME.match(value):
            raise exc.ValidationError("Invalid name.")

        return value or False

    @validates("display")
    def validate_display(self, key, value):
        if self.required:
            return True
        return value

    @validates("resource_name")
    def validate_resource_name(self, key, value):
        if value not in VALID_ATTRIBUTE_RESOURCES:
            raise exc.ValidationError("Invalid resource name.")
        return value

    @classmethod
    def all_by_name(cls, session, resource_name=None):
        query = session.query(cls)
        if resource_name is not None:
            query = query.filter_by(resource_name=resource_name)

        return {
            attribute.name: attribute
            for attribute in query.all()
        }

    @property
    def constraints(self):
        constraints = {}
        if self._constraints is not None:
            constraints = json.loads(self._constraints)

        return {
            "allow_empty": constraints.get("allow_empty", False),
            "pattern": constraints.get("pattern", ""),
            "valid_values": constraints.get("valid_values", []),
        }

    @constraints.setter
    def constraints(self, value):
        if not isinstance(value, dict):
            raise exc.ValidationError("Expected dictionary but received {}".format(
                type(value)
            ))

        constraints = {
            "allow_empty": value.get("allow_empty", False),
            "pattern": value.get("pattern", ""),
            "valid_values": value.get("valid_values", []),
        }

        if not isinstance(constraints["allow_empty"], bool):
            raise exc.ValidationError("allow_empty expected type bool.")

        if not isinstance(constraints["pattern"], basestring):
            raise exc.ValidationError("pattern expected type string.")

        if not isinstance(constraints["valid_values"], list):
            raise exc.ValidationError("valid_values expected type list")

        self._constraints = json.dumps(constraints)

    def _validate_single_value(self, value, constraints=None):
        if not isinstance(value, basestring):
            raise exc.ValidationError("Attribute values must be a string type")

        if constraints is None:
            constraints = self.constraints

        allow_empty = constraints.get("allow_empty", False)
        if not allow_empty and not value:
            raise exc.ValidationError(
                "Attribute {} doesn't allow empty values".format(self.name)
            )

        pattern = constraints.get("pattern")
        if pattern and not re.match(pattern, value):
            raise exc.ValidationError(
                "Attribute value {} for {} didn't match pattern: {}"
                .format(value, self.name, pattern)
            )

        valid_values = set(constraints.get("valid_values", []))
        if valid_values and value not in valid_values:
            raise exc.ValidationError(
                "Attribute value {} for {} not a valid value: {}"
                .format(value, self.name, ", ".join(valid_values))
            )

        return {
            "attribute_id": self.id,
            "name": self.name,
            "value": value,
        }

    def validate_value(self, value):
        if self.multi:
            if not isinstance(value, list):
                raise exc.ValidationError("Attribute values must be a list type")
        else:
            value = [value]

        inserts = []
        # This does a deserialization so save the result
        constraints = self.constraints
        for val in value:
            inserts.append(self._validate_single_value(val, constraints))

        return inserts

    def to_dict(self):
        return {
            "id": self.id,
            "site_id": self.site_id,
            "description": self.description,
            "name": self.name,
            "resource_name": self.resource_name,
            "required": self.required,
            "display": self.display,
            "multi": self.multi,
            "constraints": self.constraints,
        }


class Change(Model):
    """ Record of all changes in NSoT."""

    __tablename__ = "changes"

    id = Column(Integer, primary_key=True)

    site_id = Column(Integer, ForeignKey("sites.id", ondelete="CASCADE"), nullable=False, index=True)
    site = relationship(Site, lazy="joined", backref=backref("changes", cascade="all,delete-orphan"))

    user_id = Column(Integer, ForeignKey("users.id"), nullable=False, index=True)
    user = relationship(User, lazy="joined", backref="changes")

    change_at = Column(DateTime, default=datetime.utcnow, nullable=False)

    event = Column(String(10), nullable=False)

    resource_name = Column(String(20), nullable=False, index=True)
    resource_id = Column(Integer, nullable=False)

    _resource = Column("resource", Text, nullable=False)

    @validates("event")
    def validate_event(self, key, value):
        if value not in CHANGE_EVENTS:
            raise exc.ValidationError("Invalid Change Event.")
        return value

    @validates("resource_name")
    def validate_resource_name(self, key, value):
        if value not in VALID_CHANGE_RESOURCES:
            raise exc.ValidationError("Invalid resource name.")
        return value

    @property
    def resource(self):
        return json.loads(self._resource)

    @resource.setter
    def resource(self, value):
        self._resource = json.dumps(value)

    @classmethod
    def create(cls, session, user_id, event, resource):

        kwargs = {
            "site_id": resource.site_id,
            "user_id": user_id,
            "event": event,
            "resource_name": resource.model_name,
            "resource_id": resource.id,
            "resource": resource.to_dict(),
        }

        obj = cls(**kwargs).add(session)
        return obj

    def to_dict(self):
        resource = None
        if self.resource is not None:
            resource = self.resource

        return {
            "id": self.id,
            "site": self.site.to_dict(),
            "user": self.user.to_dict(),
            "change_at": timegm(self.change_at.timetuple()),
            "event": self.event,
            "resource_name": self.resource_name,
            "resource_id": self.resource_id,
            "resource": resource,
        }


class Counter(Model):

    __tablename__ = "counters"

    id = Column(Integer, primary_key=True)
    name = Column(String(length=64), unique=True, nullable=False)
    count = Column(Integer, nullable=False, default=0)
    last_modified = Column(DateTime, default=datetime.utcnow, nullable=False)

    @classmethod
    def incr(cls, session, name, count=1):
        counter = session.query(cls).filter_by(name=name).scalar()
        if counter is None:
            counter = cls(name=name, count=count).add(session)
            session.flush()
            return counter
        counter.count = cls.count + count
        session.flush()
        return counter

    @classmethod
    def decr(cls, session, name, count=1):
        return cls.incr(session, name, -count)



"""
Regression tests for Network objects.
"""

import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client,
    load_json, run_set_queries
)


def test_network_bug_issues_34(tornado_server):
    """Test set queries for Networks."""
    client = Client(tornado_server)

    client.create('/sites', name='Test Site')  # 1

    # Pre-load the attributes
    attr_data = load_json('attributes.json')
    client.create(
        '/sites/1/attributes',
        attributes=attr_data['attributes']
    )

    # Populate the device objects.
    network_data = load_json('networks.json')
    client.create(
        '/sites/1/networks',
        networks=network_data['networks']
    )

    # Filter networks w/ attribute hostname=foo-bar1
    n1_output = load_json('networks/bug_issues_34_1.json')
    assert_success(
        client.get("/sites/1/networks?attributes=hostname=foo-bar1"),
        n1_output['data'],
    )

    # Filter networks w/ attribute hostname=foo-bar1, including IPs
    n2_output = load_json('networks/bug_issues_34_2.json')
    assert_success(
        client.get("/sites/1/networks?attributes=hostname=foo-bar1&include_ips=True"),
        n2_output['data'],
    )



import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client
)


def test_malformed(tornado_server):
    client = Client(tornado_server)
    assert_error(client.post("/sites", user="admin", data="Non-JSON"), 400)


def test_creation(tornado_server):
    client = Client(tornado_server)
    assert_success(client.get("/sites"), {
        "sites": [],
        "limit": None,
        "offset": 0,
        "total": 0,
    })

    assert_created(client.create("/sites", name="Test Site"), "/api/sites/1")
    assert_error(client.create("/sites", name="Test Site"), 409)

    assert_success(
        client.get("/sites"),
        {
            "sites": [{"description": "", "id": 1, "name": "Test Site"}],
            "limit": None,
            "offset": 0,
            "total": 1,
        }
    )

    assert_success(
        client.get("/sites/1"),
        {"site": {"description": "", "id": 1, "name": "Test Site"}}
    )

    assert_created(client.create("/sites", name="Test Site 2"), "/api/sites/2")
    assert_success(
        client.get("/sites", params={"name": "Test Site"}),
        {
            "sites": [{"description": "", "id": 1, "name": "Test Site"}],
            "limit": None,
            "offset": 0,
            "total": 1,
        }
    )


def test_update(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")

    assert_success(
        client.update("/sites/1", name="Test Site 2"),
        {"site": {"description": "", "id": 1, "name": "Test Site 2"}}
    )

    assert_success(
        client.update("/sites/1", name="Test Site", description="A description."),
        {"site": {"description": "A description.", "id": 1, "name": "Test Site"}}
    )

    assert_error(client.update("/sites/1", description="Only description."), 400)
    assert_error(client.update("/sites/1"), 400)


def test_deletion(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")
    client.create("/sites/1/attributes", resource_name="Network", name="attr1")
    client.create("/sites/1/networks",
        cidr="10.0.0.0/24", attributes={"attr1": "foo"}
    )

    # Don't allow delete when there's an attached network/attribute
    assert_error(client.delete("/sites/1"), 409)

    client.delete("/sites/1/networks/1")

    # Stile Don't allow delete when there's an attached attribute
    assert_error(client.delete("/sites/1"), 409)

    client.delete("/sites/1/attributes/1")

    assert_deleted(client.delete("/sites/1"))



"""empty message

Revision ID: e648e9e1adb
Revises: 
Create Date: 2015-02-24 17:27:56.180895

"""

# revision identifiers, used by Alembic.
revision = 'e648e9e1adb'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('email', sa.String(length=255), nullable=False),
    sa.Column('secret_key', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('email')
    )
    op.create_table('counters',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('count', sa.Integer(), nullable=False),
    sa.Column('last_modified', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    op.create_table('sites',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('description', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    op.create_table('attributes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('site_id', sa.Integer(), nullable=False),
    sa.Column('resource_name', sa.String(length=20), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('description', sa.Text(), nullable=False),
    sa.Column('required', sa.Boolean(), nullable=False),
    sa.Column('display', sa.Boolean(), nullable=False),
    sa.Column('multi', sa.Boolean(), nullable=False),
    sa.Column('constraints', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['site_id'], [u'sites.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_attributes_resource_name'), 'attributes', ['resource_name'], unique=False)
    op.create_index('name_idx', 'attributes', ['site_id', 'resource_name', 'name'], unique=True)
    op.create_table('devices',
    sa.Column('attributes', sa.Text(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('site_id', sa.Integer(), nullable=False),
    sa.Column('hostname', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['site_id'], [u'sites.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('hostname_idx', 'devices', ['site_id', 'hostname'], unique=True)
    op.create_index(op.f('ix_devices_hostname'), 'devices', ['hostname'], unique=False)
    op.create_index(op.f('ix_devices_site_id'), 'devices', ['site_id'], unique=False)
    op.create_table('permissions',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('site_id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('permissions_flag', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['site_id'], [u'sites.id'], ondelete=u'CASCADE'),
    sa.ForeignKeyConstraint(['user_id'], [u'users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('site_user_idx', 'permissions', ['site_id', 'user_id'], unique=True)
    op.create_table('networks',
    sa.Column('attributes', sa.Text(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('site_id', sa.Integer(), nullable=False),
    sa.Column('ip_version', sa.String(length=1), nullable=False),
    sa.Column('parent_id', sa.Integer(), nullable=True),
    sa.Column('network_address', sa.VARBINARY(length=16), nullable=False),
    sa.Column('broadcast_address', sa.VARBINARY(length=16), nullable=False),
    sa.Column('prefix_length', sa.Integer(), nullable=False),
    sa.Column('is_ip', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['parent_id'], [u'networks.id'], ),
    sa.ForeignKeyConstraint(['site_id'], [u'sites.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('cidr_idx', 'networks', ['site_id', 'ip_version', 'network_address', 'prefix_length'], unique=True)
    op.create_index(op.f('ix_networks_broadcast_address'), 'networks', ['broadcast_address'], unique=False)
    op.create_index(op.f('ix_networks_ip_version'), 'networks', ['ip_version'], unique=False)
    op.create_index(op.f('ix_networks_is_ip'), 'networks', ['is_ip'], unique=False)
    op.create_index(op.f('ix_networks_network_address'), 'networks', ['network_address'], unique=False)
    op.create_index(op.f('ix_networks_prefix_length'), 'networks', ['prefix_length'], unique=False)
    op.create_index(op.f('ix_networks_site_id'), 'networks', ['site_id'], unique=False)
    op.create_table('changes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('site_id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('change_at', sa.DateTime(), nullable=False),
    sa.Column('event', sa.String(length=10), nullable=False),
    sa.Column('resource_name', sa.String(length=20), nullable=False),
    sa.Column('resource_id', sa.Integer(), nullable=False),
    sa.Column('resource', sa.Text(), nullable=False),
    sa.ForeignKeyConstraint(['site_id'], [u'sites.id'], ondelete=u'CASCADE'),
    sa.ForeignKeyConstraint(['user_id'], [u'users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_changes_resource_name'), 'changes', ['resource_name'], unique=False)
    op.create_index(op.f('ix_changes_site_id'), 'changes', ['site_id'], unique=False)
    op.create_index(op.f('ix_changes_user_id'), 'changes', ['user_id'], unique=False)
    op.create_table('network_attribute_index',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('value', sa.String(length=255), nullable=False),
    sa.Column('resource_id', sa.Integer(), nullable=False),
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['attribute_id'], [u'attributes.id'], ),
    sa.ForeignKeyConstraint(['resource_id'], [u'networks.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_network_attribute_index_name'), 'network_attribute_index', ['name'], unique=False)
    op.create_index(op.f('ix_network_attribute_index_value'), 'network_attribute_index', ['value'], unique=False)
    op.create_table('device_attribute_index',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('value', sa.String(length=255), nullable=False),
    sa.Column('resource_id', sa.Integer(), nullable=False),
    sa.Column('attribute_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['attribute_id'], [u'attributes.id'], ),
    sa.ForeignKeyConstraint(['resource_id'], [u'devices.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_device_attribute_index_name'), 'device_attribute_index', ['name'], unique=False)
    op.create_index(op.f('ix_device_attribute_index_value'), 'device_attribute_index', ['value'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_device_attribute_index_value'), table_name='device_attribute_index')
    op.drop_index(op.f('ix_device_attribute_index_name'), table_name='device_attribute_index')
    op.drop_table('device_attribute_index')
    op.drop_index(op.f('ix_network_attribute_index_value'), table_name='network_attribute_index')
    op.drop_index(op.f('ix_network_attribute_index_name'), table_name='network_attribute_index')
    op.drop_table('network_attribute_index')
    op.drop_index(op.f('ix_changes_user_id'), table_name='changes')
    op.drop_index(op.f('ix_changes_site_id'), table_name='changes')
    op.drop_index(op.f('ix_changes_resource_name'), table_name='changes')
    op.drop_table('changes')
    op.drop_index(op.f('ix_networks_site_id'), table_name='networks')
    op.drop_index(op.f('ix_networks_prefix_length'), table_name='networks')
    op.drop_index(op.f('ix_networks_network_address'), table_name='networks')
    op.drop_index(op.f('ix_networks_is_ip'), table_name='networks')
    op.drop_index(op.f('ix_networks_ip_version'), table_name='networks')
    op.drop_index(op.f('ix_networks_broadcast_address'), table_name='networks')
    op.drop_index('cidr_idx', table_name='networks')
    op.drop_table('networks')
    op.drop_index('site_user_idx', table_name='permissions')
    op.drop_table('permissions')
    op.drop_index(op.f('ix_devices_site_id'), table_name='devices')
    op.drop_index(op.f('ix_devices_hostname'), table_name='devices')
    op.drop_index('hostname_idx', table_name='devices')
    op.drop_table('devices')
    op.drop_index('name_idx', table_name='attributes')
    op.drop_index(op.f('ix_attributes_resource_name'), table_name='attributes')
    op.drop_table('attributes')
    op.drop_table('sites')
    op.drop_table('counters')
    op.drop_table('users')
    ### end Alembic commands ###



import pytest

from nsot import exc
from nsot import models

from .fixtures import session, site, user, admin


def test_networks_creation_reparenting(session, admin, site):
    net_8  = models.Network.create(session, admin.id, site.id, u"10.0.0.0/8")
    net_24 = models.Network.create(session, admin.id, site.id, u"10.0.0.0/24")
    net_16 = models.Network.create(session, admin.id, site.id, u"10.0.0.0/16")
    net_0  = models.Network.create(session, admin.id, site.id, u"0.0.0.0/0")

    assert net_0.parent_id is None
    assert net_8.parent_id == net_0.id
    assert net_16.parent_id == net_8.id
    assert net_24.parent_id == net_16.id

    assert sorted(net_0.supernets(session)) == sorted([])
    assert sorted(net_0.subnets(session)) == sorted([net_8, net_16, net_24])

    assert sorted(net_8.supernets(session)) == sorted([net_0])
    assert sorted(net_8.subnets(session)) == sorted([net_16, net_24])

    assert sorted(net_16.supernets(session)) == sorted([net_0, net_8])
    assert sorted(net_16.subnets(session)) == sorted([net_24])

    assert sorted(net_24.supernets(session)) == sorted([net_0, net_8, net_16])
    assert sorted(net_24.subnets(session)) == sorted([])


def test_seqential_creation(session, admin, site):
    net_8 = models.Network.create(session, admin.id, site.id, u"10.0.0.0/8")
    net_22_1 = models.Network.create(session, admin.id, site.id, u"10.0.0.0/22")
    net_22_2 = models.Network.create(session, admin.id, site.id, u"10.0.4.0/22")
    net_22_3 = models.Network.create(session, admin.id, site.id, u"10.0.8.0/22")

    assert net_8.id == net_22_1.parent_id
    assert net_8.id == net_22_2.parent_id
    assert net_8.id == net_22_3.parent_id


def test_network_create_hostbits_set(session, admin, site):
    with pytest.raises(ValueError):
        models.Network.create(session, admin.id, site.id, u"10.0.0.0/0")


def test_network_attributes(session, admin, site):
    models.Attribute.create(
        session, admin.id, site_id=site.id,
        resource_name="Network", name="vlan"
    )

    network = models.Network.create(session, admin.id, site.id, u"10.0.0.0/8", {
        "vlan": "34"
    })

    assert network.get_attributes() == {"vlan": "34"}

    # Verify property successfully zeros out attributes
    network.update(admin.id, attributes={})
    assert network.get_attributes() == {}

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes=None)

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={0: "value"})

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"key": 0})

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"made_up": "value"})


def test_ip_address_no_network(session, admin, site):
    with pytest.raises(exc.ValidationError):
        models.Network.create(session, admin.id, site.id, u"10.0.0.1/32")

    models.Network.create(session, admin.id, site.id, u"10.0.0.0/8")
    models.Network.create(session, admin.id, site.id, u"10.0.0.1/32")

def test_retrieve_networks(session, admin, site):
    models.Attribute.create(
        session, admin.id, site_id=site.id,
        resource_name="Network", name="test"
    )

    net_8 = models.Network.create(
        session, admin.id, site.id, cidr=u"10.0.0.0/8",
        attributes={"test": "foo"}
    )
    net_24 = models.Network.create(
        session, admin.id, site.id, cidr=u"10.0.0.0/24",
        attributes={"test": "bar"}
    )
    ip = models.Network.create(
        session, admin.id, site.id, cidr=u"10.0.0.1/32"
    )

    assert sorted(site.networks(
        root=True
    )) == sorted([net_8])

    assert sorted(site.networks(
        include_networks=True, include_ips=False
    )) == sorted([net_8, net_24])

    assert sorted(site.networks(
        include_networks=False, include_ips=False
    )) == sorted([])

    assert sorted(site.networks(
        include_networks=True, include_ips=True
    )) == sorted([net_8, net_24, ip])

    assert sorted(site.networks(
        include_networks=False, include_ips=True
    )) == sorted([ip])

    assert sorted(site.networks(
        subnets_of="10.0.0.0/10"
    )) == sorted([net_24])

    assert sorted(site.networks(
        subnets_of="10.0.0.0/10", include_ips=True
    )) == sorted([net_24, ip])

    assert sorted(site.networks(
        supernets_of="10.0.0.0/10"
    )) == sorted([net_8])

    with pytest.raises(ValueError):
        site.networks(subnets_of="10.0.0.0/10", supernets_of="10.0.0.0/10")

    with pytest.raises(ValueError):
        assert site.networks(attribute_value="foo")

    assert sorted(site.networks(
        include_ips=True, attribute_name="test"
    )) == sorted([net_8, net_24])

    assert sorted(site.networks(
        include_ips=True, attribute_name="test", attribute_value="foo"
    )) == sorted([net_8])



import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client,
    load_json, run_set_queries
)


def test_creation(tornado_server):
    admin_client = Client(tornado_server, "admin")
    user_client = Client(tornado_server, "user")

    admin_client.create("/sites", name="Test Site")  # 1
    admin_client.create(
        "/sites/1/attributes",
        resource_name="Device", name="attr1"
    )  # 1

    # Invalid permissions
    assert_error(
        user_client.create(
            "/sites/1/devices", hostname="device1", attributes={"attr1": "foo"}
        ),
        403
    )

    # Missing required field (hostname)
    assert_error(
        admin_client.create(
            "/sites/1/devices", attributes={"attr1": "foo"}
        ),
        400
    )

    # Null hostname
    assert_error(
        admin_client.create(
            "/sites/1/devices",
            hostname=None
        ),
        400
    )

    # Verify Successful Creation
    assert_created(
        admin_client.create(
            "/sites/1/devices", hostname="device1", attributes={"attr1": "foo"}
        ),
        "/api/sites/1/devices/1"
    )

    # Verify Successful get of all Devices
    assert_success(
        admin_client.get("/sites/1/devices"),
        {
            "devices": [{
                "attributes": {"attr1": "foo"},
                "id": 1,
                "site_id": 1,
                "hostname": "device1",
            }],
            "limit": None,
            "offset": 0,
            "total": 1,
        }

    )

    # Verify Successful get of single Device
    assert_success(
        admin_client.get("/sites/1/devices/1"),
        {"device": {
            "attributes": {"attr1": "foo"},
            "id": 1,
            "site_id": 1,
            "hostname": "device1",
        }}
    )


def test_collection_creation(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")  # 1

    # Successfully create a collection of Devices
    collection = [
        {"hostname": "device1"},
        {"hostname": "device2"},
        {"hostname": "device3"},
    ]
    collection_response = client.create(
        "/sites/1/devices",
        devices=collection
    )
    assert_created(collection_response, None)

    # Successfully get all created Devices
    output = collection_response.json()
    output['data'].update({"limit": None, "offset": 0})

    assert_success(
        client.get("/sites/1/devices"),
        output['data'],
    )


def test_filters(tornado_server):
    """Test hostname/attribute filters for Devices."""
    client = Client(tornado_server)

    client.create('/sites', name='Test Site')  # 1

    # Pre-load the attributes
    attr_data = load_json('attributes.json')
    client.create(
        '/sites/1/attributes',
        attributes=attr_data['attributes']
    )

    # Populate the device objects.
    device_data = load_json('devices.json')
    client.create(
        '/sites/1/devices',
        devices=device_data['devices']
    )

    # Test lookup by hostnme
    hostname_output = load_json('devices/filter1.json')
    assert_success(
        client.get("/sites/1/devices?hostname=foo-bar3"),
        hostname_output['data'],
    )

    # Test lookup by attributes
    attr_output = load_json('devices/filter2.json')
    assert_success(
        client.get("/sites/1/devices?attributes=foo=baz"),
        attr_output['data'],
    )

    # Test lookup with multiple attributes
    multiattr_output = load_json('devices/filter3.json')
    assert_success(
        client.get("/sites/1/devices?attributes=foo=baz&attributes=cluster=lax"),
        multiattr_output['data'],
    )


def test_set_queries(tornado_server):
    """Test set queries for Devices."""
    client = Client(tornado_server)

    client.create('/sites', name='Test Site')  # 1

    # Pre-load the attributes
    attr_data = load_json('attributes.json')
    client.create(
        '/sites/1/attributes',
        attributes=attr_data['attributes']
    )

    # Populate the device objects.
    device_data = load_json('devices.json')
    client.create(
        '/sites/1/devices',
        devices=device_data['devices']
    )

    # Mapping of query string to file containing expected response data for
    # each query.
    device_queries = (
        # INTERSECTION: foo=bar
        ('foo=bar', 'query1.json'),
        # INTERSECTION: foo=bar owner=jathan
        ('foo=bar owner=jathan', 'query2.json'),
        # DIFFERENCE: -owner=gary
        ('-owner=gary', 'query3.json'),
        # UNION: cluster +foo=baz
        ('cluster +foo=baz', 'query4.json'),
    )
    run_set_queries('devices', client, device_queries)


def test_update(tornado_server):
    admin_client = Client(tornado_server, "admin")
    user_client = Client(tornado_server, "user")

    admin_client.create("/sites", name="Test Site")  # 1
    admin_client.create(
        "/sites/1/attributes", resource_name="Device", name="attr1"
    )  # 1
    admin_client.create(
        "/sites/1/devices", hostname="foo", attributes={"attr1": "foo"}
    )

    # Empty Update should only clear attributes.
    assert_success(
        admin_client.update("/sites/1/devices/1", hostname="foo"),
        {"device": {
            "attributes": {},
            "id": 1,
            "site_id": 1,
            "hostname": "foo",
        }}
    )

    # Now put attributes back and change hostname
    assert_success(
        admin_client.update(
            "/sites/1/devices/1", hostname="bar", attributes={"attr1": "foo"}
        ),
        {"device": {
            "attributes": {"attr1": "foo"},
            "id": 1,
            "site_id": 1,
            "hostname": "bar",
        }}
    )

    # Invalid permissions
    assert_error(
        user_client.update("/sites/1/devices/1"),
        403
    )


def test_deletion(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")

    client.create("/sites/1/devices", hostname="device1")  # 1
    client.create("/sites/1/devices", hostname="device2")  # 2

    assert_deleted(client.delete("/sites/1/devices/1"))



import functools

from . import exc
from .permissions import PermissionsFlag


def _has_any_perm(user, site_id, perms):
    permissions = set(
        user.get_permissions(site_id)
            .get(site_id, {})
            .get("permissions", [])
    )
    for perm in perms:
        if perm in permissions:
            return True
    return False


def any_perm(*perms):
    """ Takes a list of permissions and fails if none are valid for a user.

    """
    def dec(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if "site_id" not in kwargs:
                raise ValueError(
                    "Permission decorator used on handler method "
                    "without site_id keyword argument."
                )
            if not _has_any_perm(self.current_user, kwargs["site_id"], perms):
                raise exc.Forbidden("Lacking appropriate permissions.")
            return method(self, *args, **kwargs)
        return wrapper
    return dec



import pytest

from nsot import exc
from nsot import models

from .fixtures import session, site, user, admin


def test_device_attributes(session, admin, site):
    models.Attribute.create(
        session, admin.id, site_id=site.id,
        resource_name="Device", name="owner"
    )

    device = models.Device.create(session, admin.id, site.id, "foobarhost", {
        "owner": "gary"
    })

    assert device.get_attributes() == {"owner": "gary"}

    # Verify property successfully zeros out attributes
    device.update(admin.id, attributes={})
    assert device.get_attributes() == {}

    with pytest.raises(exc.ValidationError):
        device.update(admin.id, attributes=None)

    with pytest.raises(exc.ValidationError):
        device.update(admin.id, attributes={0: "value"})

    with pytest.raises(exc.ValidationError):
        device.update(admin.id, attributes={"key": 0})

    with pytest.raises(exc.ValidationError):
        device.update(admin.id, attributes={"made_up": "value"})


def test_retrieve_device(session, admin, site):
    models.Attribute.create(
        session, admin.id, site_id=site.id,
        resource_name="Device", name="test"
    )

    device1 = models.Device.create(
        session, admin.id, site.id, hostname="device1",
        attributes={"test": "foo"}
    )
    device2 = models.Device.create(
        session, admin.id, site.id, hostname="device2",
        attributes={"test": "bar"}
    )
    device3 = models.Device.create(
        session, admin.id, site.id, hostname="device3"
    )

    assert sorted(site.devices()) == sorted([device1, device2, device3])

    with pytest.raises(ValueError):
        assert site.devices(attribute_value="foo")

    assert sorted(site.devices(
        attribute_name="test"
    )) == sorted([device1, device2])

    assert sorted(site.devices(
        attribute_name="test", attribute_value="foo"
    )) == sorted([device1])
